=begin rdoc

==scriptingutils - some methods useful when writing scripts

Scripts using this module should first call the +init+ method.

=== Copyright

Copyright (C) 2009 Wybo Dekker (<tt>wybo@servalys.nl</tt>)

This program is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but without any warranty; without even the implied warranty of
merchantability or fitness for a particular purpose. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see http://www.gnu.org/licenses.

=end

require 'rubygems'
require 'optparse'

[ 'rio', 'session', 'term/ansicolor'].each do |g|
  begin
    require g
  rescue LoadError
    puts "You probably have not installed the #{g} gem!"
    exit 1
  end
end

   MYNAME = File.basename($0)
  WORKDIR = ENV['PWD'] # don't use Dir.getwd - this one is the real workdir
                       # even if called from mk's tempdir
  PATHSEP = File::PATH_SEPARATOR

   TONULL = ' 2> /dev/null'
 FROMNULL = ' < /dev/null'
   DEVTTY = '/dev/tty'
  PRINTER = 'lpr'

class String
  # include coloring methods for strings: red, green, bold et cetera:
  include Term::ANSIColor

  # escape tex specials with a backslash
  # example:
  # 'tool_example'.texescape -> 'tool\_example'
  def texescape 
    self.gsub(/([#&$_{}%])/) { '\\'+$1 }.gsub(/([<>])/,'$\1$')
  end
end

class Array
  # escape tex specials in String-type elements with a backslash
  # example:
  # ['tool_example',12].texescape -> ['tool\_example',12]
  def texescape
    self.map { |x| x.class == String ? x.texescape : x }
  end
end

# return prog name, output, success rate, and exit status of a system command
# :call-seq:
#   h = sys(command)
#   h[:progname] -> name of the external program being run
#   h[:output]   -> output of the program on stdout
#   h[:error]    -> output of the program on stderr
#   h[:succes]   -> true if the program ran successfully, false otherwise
#   h[:status]   -> exit status of the program ($?)
def sys(command,interactive=false)
  quit('sys called with nil or empty string',1) if (command||'').empty? 
  prog = out = err = success = status = nil
  if interactive
    system(command)
    status = $?.exitstatus
  else
    sh = Session::Shell.new
    # escape spaces inside " or ' with ###:
    #!! later: ook tabs afvangen
    com = command.gsub(/(['"]).*?\1/) { |x| x.gsub(/ /,'###') }.split
    while com.size > 0
      prog = com.shift.gsub(/###/,' ')
      if prog =~ /=/ # leading envvar setting
        sh.execute("export #{prog}")
      else
        com.unshift(prog)
        out,err = sh.execute(com.join(' '))
        break
      end
    end
    status = sh.status
    sh.close
  end
  success = status == 0
  
  mess = "system call: ".yellow.bold + command.cyan
  if @verbose
    if success
      warn mess + " successful"
    else
      warn mess + " #{'not successful'.red} exit status=#{status.to_s.red}"
      warn <<-EOF
        progname:   #{prog}
        success:    false
        status:     #{status}
        output:     #{(out||'').gsub(/\n/m,"\n                    ")}
        error:      #{err} 
      EOF
    end
  end
  {:progname => prog, :output => out, :error => err, :success => success, :status => status}
end

# executable returns true if an executable _ex_ exists in the PATH;
# if it does not, it returns false, or it dies, depending on _needed_.
def executable(ex,needed=false)
  ex.strip!
  ex =~ /\// and quit("executable (#{ex}) should not contain a /".red.bold,1)
  ENV['PATH'].split(PATHSEP).each do |dir|
    f = File.join(dir,ex)
    return true if File.exist?(f) && File.executable?(f)
  end
  mess = "executable #{ex} needed but not found or not executable".red.bold
  quit(mess,1) if needed
  warn(mess)
  return false
end

# print version and quit
def print_version(release)
  v,d = []
  open($0).each do |$_| 
    v,d = $1,$2  if ~ /\$Id: .* ([\d.]+) (\d+-\d+-\d+)/
  end
  quit("#{MYNAME}: release #{release}, CVS-version #{v} (#{d})")
end

# read an rc file
def read_rc_file(rc,must_exist=false)
  unless rc && File.exist?(rc) && File.readable?(rc)
    quit("rc file #{rc} does not exist or is not readable".red.bold,1) if must_exist
    return
  end
  begin
    open(rc) { |o| eval o.read }
  rescue
    quit("Errors handling rc file #{rc}: #{$!}".red.bold,1)
  end
  @rcfiles.push(rc) # remember which files were read
end

# program initializer: 
# 1. set the defaults (values of _defaults_) for
#    option- and rc-variables (keys of _defaults_)
# 2. read the rc-files
# 3. handle the options, as defined in the _optionproc_ method
# 4. if the --rc option was used, read its arg as an rc-file
# 5. check if all needed executables (in _needed_, maybe nil) are available
def init(needed,optionproc)
  @rcfiles = [] # collect names for a report in case @verbose is true
  @verbose = false
  @starting_dir = Dir.pwd

  # run all standard rc-files
  unless ENV['NORC']
    [ "/etc/#{MYNAME}rc",               # system rc file
      "#{ENV['HOME']}/.#{MYNAME}rc",    # user rc file
      ".#{MYNAME}rc",                   # local rc file
    ].each do |rc|
      read_rc_file(rc)
    end
  end

  # handle options
  help = method(optionproc).call

  # in case a --rc option was used, read its argument as an rc-file:
  read_rc_file(@rc,true) if @rc

  # if the --verbose option was used, report rc-files read:
  if @verbose
    if @rcfiles.size > 0
      warn "These rc files were read:\n   #{@rcfiles.join("\n   ")}"
    else
      warn "No rc files were read"
    end
  end

  # check if all needed executables (in _needed_) are available
  check_execs(needed)
  help
end

# - print a newline, if newline is true, 
# - print the message if not nil, 
# - then exit with the exitvalue
# - print a traceback when exitvalue is between 1 and 10 (reserving higher value for communication between processes)
def quit(mess=nil,exitvalue=0,newline=false)
  warn '' if newline # ensure prompt at start of line
  warn mess if mess
  if exitvalue.between?(1,10)
    for c in caller(0)
      warn c.sub(/#{$0}/,MYNAME) if c =~ /#{MYNAME}/
    end
  end
  # just in case we have chdir-ed into a temporary directory
  # that may have appeared overnight...
  Dir.chdir(@starting_dir)
  exit(exitvalue)
end

# check the presence of a list of executables
def check_execs(needs)
  needed = case needs
    when Array  then needs
    when String then needs.split
    when Symbol then send(needs)
    when NilClass then []
    else 
      nil
  end
  needed or quit("unsupported class #{needs.class} for needs in check_execs".red.bold,1)
  missing = []
  needed.each do |n|
    n = n.split.shift
    executable(n) or missing.push(n)
  end
  plur = missing.size > 1 ? 's' : ''
  quit("Missing executable#{plur} #{missing.join(" ")}".red.bold,1) if missing.size > 0
end

# extract a TeX log file, showing the most probable error in red and 
# the error's line number in green.
#
# with _web_ == true, produce html, (not yet implemented)
# otherwise produce ANSI
#
# return with:
# - file in which the error was found, plus 
# - the linenumber in that file, and
# - a hash with a key for each file containing an error and/or warnings.
#   the value is an array containing the lines describing the errors/warnings
#   for the file represented by the key.

def texlog_extract(file,warnings_to_skip=nil,web=false)
  quit("web coloring for texlog_extract not yet implemented!") if web
  warnings_to_skip ||= []
  # read all lines in array log, chomp the lines, and add an empty line 
  if File.exist?(file)
    log = open(file).readlines.map { |x| x.chomp }.push("")
  else
    STDERR.puts("No log file (#{file})")
    return
  end           
  
  message = {}
  currentfile = ['']
  message[''] = []
  linenum = 0
  if log[0] !~ /^This is [a-zA-Z]*TeX, Version/
    return nil, 0, { file => "This is probably not a TeX log file".red.bold }
  end
  # undo linebreaks at 80-char lengths
  for i in (0...log.size) do 
    # a line reporting the line number of the error may be preceded
    # by an 80-char line, without being its continuation:
    if log[i].size == 79 && log[i+1] !~ /l\.(\d+)/
      log[i] << log[i+1]
      log[i+1] = ''
    end
  end
  content = ''
  nwarn = p = nerr = 0
  while line = log.shift
    case line
    when /^\s*$/ then next
    when /^(Here is|Try typing)/ then break
    when /\(Font\)/ then next
    when /^LaTeX Font Info/ then next
    when /^File:/ then next
    when /warning:\s+/i
      skip = false
      warnings_to_skip.each do |w|
        if line =~ /#{w}/
          skip = true
          break
        end
      end
      unless skip
        nwarn += 1
        message[currentfile[-1]].push(line.sub(/warning:\s+/i,'\&'.magenta))
        continue = true
        packagename = line =~ /^Package (.*) warning/i ? $1 : 'unknown'
        # some warnings are followed by explanation lines, starting
        # either with whitespace or with the package name in
        # parenthesis+at least 2 spaces:
        while continue
           l = log.shift
           if l =~ /^(\((#{packagename})\))?\s\s+/
             message[currentfile[-1]].push(l.sub(/^(\(.*?\))/) {$1.tr('[(.)A-Za-z]',' ')})
           else 
             continue = false
             log.unshift(l)
           end
        end
      end
    when /^No pages of output/
      nerr += 1
      message[currentfile[-1]].push(line)
    end
  
    # keep track of the file we are in:     
    unless p > 0 # but only up to the first error
      # remove 1: () and 2: things between ()
      line.sub(/\(\)/,'').gsub(/\(\S+?\)/,'').scan(/\(\S+|\)/).each do |f|
        if f == ')'
          currentfile.pop
        else
          f = f.slice(1..-1) # without the initial (
          currentfile.push f
          message[f] ||= []
        end
      end
    end

    # recent texi2dvi uses TeX-option --file-line-error, which
    # replaces the ! with <filename>:<lineno>:
    if line =~ /^!/ or line =~ /^#{currentfile[-1]}:/
      nerr += 1
      if linenum > 0 # this is the second error message
        p = 20
      else # first error's linenumber
        p += 1
        line = line.red
      end
    elsif line =~ /^l\.(\d+) (.*)/ && p > 0 # look for a line number only after an error
      linenum = $1.to_i
      content = $2
      line = line.green
    end
  
    if p > 0
      message[currentfile[-1]].push(line)
      break if (p += 1) > 20 # this must be enough to see what's wrong
    end
  end
  
  if currentfile[-1] and currentfile[-1] != ''
    message[currentfile[-1]].push('file'.red + ': ' + currentfile[-1])
    message[currentfile[-1]].push('line'.red + " #{linenum}: #{content}")
  end
  message[''] ||= []
  message[''].push("#{nerr} errors #{nwarn} warnings")

  # check if there is a bibtex log (blg) file and add it if so:
  blg = file.sub(/\.log$/,'.blg')
  if File.exist?(blg)
    m = "From bibtex log file:\n".red # used only if there are errors
    pr = false # start printing only after a Database line
    bib = nil
    open(blg).readlines.each do |line|
      if line =~ /^Database file.*: (.*)/
        # if we saw already messages for the first .bib file, report only those
        break if pr && !message[bib].empty?
        bib = $1
        currentfile.push(bib)
        message[bib] = []
        pr = true
        next
      end
      break if line =~ /^You've used/
      if pr
        message[currentfile[-1]].push(m+line.sub(/^Illegal/,'!\&'))
        m = ''
      end
    end
  end
  message.delete_if { |k,v| v == [] }
  return currentfile.pop,linenum,(nerr == 0 && nwarn == 0) ? [] : message
end # texlog_extract
